// Author: A.Voss@FH-Aachen.de
//
// C-macros, predefined macros
// tags: #define, #undef, #if, #else, #endif

#include <stdio.h>
#include <stdlib.h>
#include "../../include/tools.h"

int main()
{
    tools_log(); printf("C-macros \n");
    
    // as convention a macro is written capitalized

    // the preprocessor replaces a macro with its definition
    #define PI 3.1415926
    tools_log(); printf("  PI = %f \n",PI);

    #define NAME "Me"
    tools_log(); printf("  Name = %s \n",NAME);
    
    // it can contain other macros, e.g. predefined ones
    #define LINE __LINE__
    tools_log(); printf("  code line = %d \n",LINE);
    
    // it may also look like a function
    #define MAX(a,b) (a<b) ? b : a
    tools_log(); printf("  max(3,5)? %d \n",MAX(3,5));
    tools_log(); printf("  max(5,3)? %d \n",MAX(5,3));

    // or
    #define MUL(a,b) a*b
    tools_log(); printf("  mul(3,5)? %d \n",MUL(3,5));

    // there are some problems, e.g.
    tools_log(); printf("  mul(1+2,5)? %d ??? \n",MUL(1+2,5));

    // or
    int i=3,j=4;
    tools_log(); printf("  i=%d j=%d",i,j);
    printf(" -> max(++i,++j)? %d ",MAX(++i,++j));
    printf(" -> i=%d j=%d ??? \n",i,j);
    
    // one solution is to embrace all (!) arguments, i.e.
    #undef MUL
    #define MUL(a,b) (a)*(b)
    tools_log(); printf("  mul(1+2,5)? %d !!! \n\n",MUL(1+2,5));
    
    // sometimes we have to distinguish between different versions or compilers
    // see "Common Predefined Macros"
    // note: the condition is evaluated at compile time, i.e. even syntax errors 
    // are possible in the other branch because it is never compiled
    #if defined(__GNUC__)
    tools_log(); printf("gnu compiler, vers. %s \n",__VERSION__);
    #else
    tools_log(); printf("no gnu compiler \n");
    #endif

    // hint: avoid macros if possible, write functions instead!
    // this also ensures type safety!
    // in rare cases macros cannot be avoided, e.g. __LINE__, there
    // is no other solution, but in general avoid them!

    return EXIT_SUCCESS;
}

